#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           Var,
                           units)
from pyomo.util.check_units import (assert_units_consistent,
                                    assert_units_equivalent)
from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)
from idaes.generic_models.unit_models.plug_flow_reactor import PFR

from idaes.generic_models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock)
from idaes.generic_models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock)
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables,
                                              number_derivative_variables)
from idaes.core.util.testing import (PhysicalParameterTestBlock,
                                     ReactionParameterTestBlock,
                                     initialization_tester)
from idaes.core.util import get_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.component
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.reactions = ReactionParameterTestBlock(default={
                            "property_package": m.fs.properties})

    m.fs.unit = PFR(default={"property_package": m.fs.properties,
                             "reaction_package": m.fs.reactions})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 19

    assert m.fs.unit.config.material_balance_type == \
        MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert not m.fs.unit.config.has_equilibrium_reactions
    assert not m.fs.unit.config.has_phase_equilibrium
    assert not m.fs.unit.config.has_heat_of_reaction
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.reaction_package is m.fs.reactions

    assert m.fs.unit.config.length_domain_set == [0.0, 1.0]
    assert m.fs.unit.config.transformation_method == "dae.finite_difference"
    assert m.fs.unit.config.transformation_scheme == "BACKWARD"
    assert m.fs.unit.config.finite_elements == 20
    assert m.fs.unit.config.collocation_points == 3


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(default={
                                "property_package": m.fs.properties})

        m.fs.unit = PFR(default={"property_package": m.fs.properties,
                                 "reaction_package": m.fs.reactions,
                                 "has_equilibrium_reactions": False,
                                 "has_heat_transfer": True,
                                 "has_heat_of_reaction": True,
                                 "has_pressure_change": True})

        m.fs.unit.inlet.flow_vol.fix(1.0)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(101325.0)

        m.fs.unit.length.fix(0.5)
        m.fs.unit.area.fix(0.1)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):
        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert isinstance(sapon.fs.unit.area, Var)
        assert isinstance(sapon.fs.unit.length, Var)
        assert isinstance(sapon.fs.unit.volume, Var)
        assert hasattr(sapon.fs.unit, "performance_eqn")
        assert hasattr(sapon.fs.unit.control_volume, "heat")
        assert hasattr(sapon.fs.unit, "heat_duty")
        assert hasattr(sapon.fs.unit, "deltaP")

        assert number_variables(sapon) == 654
        assert number_total_constraints(sapon) == 590
        assert number_unused_variables(sapon) == 14
        assert number_derivative_variables(sapon) == 0

    @pytest.mark.component
    def test_units(self, sapon):
        assert_units_consistent(sapon)
        assert_units_equivalent(sapon.fs.unit.volume, units.m**3)
        assert_units_equivalent(sapon.fs.unit.length, units.m)
        assert_units_equivalent(sapon.fs.unit.area, units.m**2)
        assert_units_equivalent(sapon.fs.unit.heat_duty[0, 0], units.W/units.m)
        assert_units_equivalent(sapon.fs.unit.deltaP[0, 0], units.Pa/units.m)

    @pytest.mark.unit
    def test_dof(self, sapon):
        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, sapon):
        assert (pytest.approx(101325.0, abs=1e-2) ==
                sapon.fs.unit.outlet.pressure[0].value)
        assert (pytest.approx(303.6, abs=1e-2) ==
                sapon.fs.unit.outlet.temperature[0].value)
        assert (pytest.approx(62.29, abs=1e-2) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, sapon):
        assert abs(value(sapon.fs.unit.inlet.flow_vol[0] -
                         sapon.fs.unit.outlet.flow_vol[0])) <= 1e-6
        assert (abs(value(sapon.fs.unit.inlet.flow_vol[0] *
                          sum(sapon.fs.unit.inlet.conc_mol_comp[0, j]
                              for j in sapon.fs.properties.component_list) -
                          sapon.fs.unit.outlet.flow_vol[0] *
                          sum(sapon.fs.unit.outlet.conc_mol_comp[0, j]
                              for j in sapon.fs.properties.component_list)))
                <= 1e-6)

        hrxn = 0
        for x in sapon.fs.unit.control_volume.length_domain:
            if x != 0:
                hrxn += value(
                    sapon.fs.unit.control_volume.heat_of_reaction[0, x] *
                    (x-sapon.fs.unit.control_volume.length_domain.prev(x)) *
                    sapon.fs.unit.control_volume.length)
        assert (pytest.approx(1847000, abs=1e3) == hrxn)
        assert abs(value(
                (sapon.fs.unit.inlet.flow_vol[0] *
                 sapon.fs.properties.dens_mol *
                 sapon.fs.properties.cp_mol *
                 (sapon.fs.unit.inlet.temperature[0] -
                    sapon.fs.properties.temperature_ref)) -
                (sapon.fs.unit.outlet.flow_vol[0] *
                 sapon.fs.properties.dens_mol *
                 sapon.fs.properties.cp_mol *
                 (sapon.fs.unit.outlet.temperature[0] -
                  sapon.fs.properties.temperature_ref)) +
                hrxn)) <= 1e-6

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, sapon):
        sapon.fs.unit.report()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_costing(self, sapon):
        sapon.fs.unit.get_costing()
        assert isinstance(sapon.fs.unit.costing.purchase_cost, Var)
        results = solver.solve(sapon)
        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok
        assert (pytest.approx(9869.51230, abs=1e3) ==
                value(sapon.fs.unit.costing.base_cost))
        assert (pytest.approx(16025.42623, abs=1e3) ==
                value(sapon.fs.unit.costing.purchase_cost))

        assert_units_consistent(sapon.fs.unit.costing)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import pyomo.common.config
import logging.config
import json
import yaml
import os
import importlib

_log = logging.getLogger(__name__)
# Default release version if no options provided for get-extensions
default_binary_release = "2.4.4"
# Where to download releases from get-extensions
release_base_url = "https://github.com/IDAES/idaes-ext/releases/download"
# Where to get release checksums
release_checksum_url = \
    "https://raw.githubusercontent.com/IDAES/idaes-ext/main/releases/sha256sum_{}.txt"
# Map some platform names to others for get-extensions
binary_platform_map = {
    "rhel7": "centos7",
    "rhel8": "centos8",
    "ubuntu1810": "ubuntu1804",
    "ubuntu1904": "ubuntu1804",
    "ubuntu1910": "ubuntu1804",
    "ubuntu2010": "ubuntu2004",
    "ubuntu2104": "ubuntu2004",
    "linux": "centos7",
}
# Set of known platforms with available binaries and descriptions of them
known_binary_platform = {
    "auto":"Auto-select windows, darwin or linux",
    "windows":"Microsoft Windows (built on verion 1909)",
    "darwin": "OSX (currently not available)",
    "linux": (
            f"Linux (auto select distribution or fall back on "
            f"{binary_platform_map['linux']})"
        ),
    "centos7": "CentOS 7",
    "centos8": "CentOS 8",
    "rhel7": "Red Hat Enterprise Linux 7",
    "rhel8": "Red Hat Enterprise Linux 8",
    "ubuntu1804": "Ubuntu 18.04",
    "ubuntu1810": "Ubuntu 18.10",
    "ubuntu1904": "Ubuntu 19.04",
    "ubuntu1910": "Ubuntu 19.10",
    "ubuntu2004": "Ubuntu 20.04",
    "ubuntu2010": "Ubuntu 20.10",
    "ubuntu2104": "Ubuntu 21.04",
}
# Unsupported platforms
unsupported_binary_platform = ["darwin"]
# Set of extra binary packages and platforms where they are available
extra_binaries = {
    "petsc": ["centos7", "centos8", "ubuntu1804", "ubuntu2004"],
}

# Store the original environment variable values so we can revert changes
orig_environ = {
    "PATH": os.environ.get("PATH", ""),
    "LD_LIBRARY_PATH": os.environ.get("LD_LIBRARY_PATH", ""),
    "DYLD_LIBRARY_PATH": os.environ.get("DYLD_LIBRARY_PATH", ""),
}

def basic_platforms():
    """Return a set of platforms that binaries should be available for.
    """
    kp = set(known_binary_platform.keys()) - set(["auto", "linux"])
    kp -= set(unsupported_binary_platform)
    kp -= set([k for k, v in binary_platform_map.items() if k != v])
    return kp

class ConfigBlockJSONEncoder(json.JSONEncoder):
    """ This class handles non-serializable objects that may appear in the IDAES
    ConfigBlock. For now this is only set objects.
    """
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return obj

def _new_idaes_config_block():
    """The idaes configuration is stored in a Pyomo ConfigBlock created by this.
    This function is called when importing IDAES.  This function should only be
    called in ``__init__.py`` for ``idaes``.  Calling it anywhere else will cause
    the idaes configuration system to function improperly.
    """
    cfg = pyomo.common.config.ConfigBlock("idaes", implicit=True)
    cfg.declare(
        "logging",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Logging configuration dictionary",
            doc="This stores the logging configuration. See the Python "
            "logging.config.dictConfig() documentation for details.",
        ),
    )
    cfg["logging"].declare(
        "version", pyomo.common.config.ConfigValue(domain=int, default=1))
    cfg["logging"].declare(
        "disable_existing_loggers",
        pyomo.common.config.ConfigValue(domain=bool, default=False))
    cfg["logging"].declare(
        "formatters", pyomo.common.config.ConfigBlock(implicit=True))
    cfg["logging"]["formatters"].declare(
        "default_format", pyomo.common.config.ConfigBlock(implicit=True))
    cfg["logging"]["formatters"]["default_format"].declare(
        "format",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="%(asctime)s [%(levelname)s] %(name)s: %(message)s"))
    cfg["logging"]["formatters"]["default_format"].declare(
        "datefmt",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="%Y-%m-%d %H:%M:%S"))
    cfg["logging"].declare(
        "handlers", pyomo.common.config.ConfigBlock(implicit=True))
    cfg["logging"]["handlers"].declare(
        "console", pyomo.common.config.ConfigBlock(implicit=True))
    cfg["logging"]["handlers"]["console"].declare(
        "class",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="logging.StreamHandler"))
    cfg["logging"]["handlers"]["console"].declare(
        "formatter",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="default_format"))
    cfg["logging"]["handlers"]["console"].declare(
        "stream",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="ext://sys.stdout"))
    cfg["logging"].declare(
        "loggers",
        pyomo.common.config.ConfigValue(
            domain=dict,
            default={ # the period in the logger names is trouble for ConfigBlock
                "idaes":{
                    "level": "INFO",
                    "propagate": True,
                    "handlers": ["console"]
                },
                "idaes.solve":{
                    "propagate": False,
                    "handlers": ["console"]
                },
                "idaes.init":{
                    "propagate": False,
                    "handlers": ["console"]
                },
                "idaes.model":{
                    "propagate":False,
                    "handlers": ["console"]
                }}))
    cfg.declare(
        "ipopt",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Default config for 'ipopt' solver",
            doc="Default config for 'ipopt-iades' solver"
        ),
    )

    cfg["ipopt"].declare(
        "options",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Default solver options for 'ipopt'",
            doc="Default solver options for 'ipopt' solver"
        ),
    )

    cfg["ipopt"]["options"].declare(
        "nlp_scaling_method",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="gradient-based",
            description="Ipopt NLP scaling method",
            doc="Ipopt NLP scaling method"
        ),
    )

    cfg["ipopt"]["options"].declare(
        "tol",
        pyomo.common.config.ConfigValue(
            domain=float,
            default=1e-6,
            description="Ipopt tol option",
            doc="Ipopt tol option"
        ),
    )

    cfg.declare(
        "default_solver",
        pyomo.common.config.ConfigValue(
            default="ipopt",
            domain=str,
            description="Default solver.  See Pyomo's SolverFactory for detauls.",
            doc="Default solver.  See Pyomo's SolverFactory for detauls.",
        ),
    )

    cfg.declare(
        "use_idaes_solvers",
        pyomo.common.config.ConfigValue(
            default=True,
            domain=bool,
            description="If True, search the IDAES bin directory for executables"
                        " first; otherwise, use IDAES bin directory as last resort.",
            doc="If True the IDAES bin directory will be searched for executables "
                "first, which will result in the solvers installed by IDAES being "
                "used in preference to solvers installed on the machine.  If False, "
                "IDAES will only fall back on solvers installed in the IDAES bin "
                "directory if they are not otherwise available.",
        ),
    )

    cfg.declare(
        "use_idaes_solver_config",
        pyomo.common.config.ConfigValue(
            default=False,
            domain=bool,
            description="If True, use the configure IDAES solver default options.",
            doc="If True, use the configure IDAES solver default options.",
        ),
    )

    cfg.declare(
        "valid_logger_tags",
        pyomo.common.config.ConfigValue(
            default=set([
                "framework",
                "model",
                "flowsheet",
                "unit",
                "control_volume",
                "properties",
                "reactions",
                "ui"]),
            domain=set,
            description="List of valid logger tags",
        ),
    )

    cfg.declare(
        "logger_tags",
        pyomo.common.config.ConfigValue(
            default=set([
                "framework",
                "model",
                "flowsheet",
                "unit",
                "control_volume",
                "properties",
                "reactions"]),
            domain=set,
            description="List of logger tags to allow",
        ),
    )

    cfg.declare(
        "logger_capture_solver",
        pyomo.common.config.ConfigValue(
            default=True,
            description="Solver output captured by logger?",
        ),
    )
    return cfg


def read_config(val, cfg):
    """Read either a JSON formatted config file or a configuration dictionary.
    Args:
        val: dict, ConfigDict, or file to read
        cfg: config block
    Returns:
        None
    """
    config_file = None
    if val is None:
        return
    elif isinstance(val, (dict, pyomo.common.config.ConfigBlock)):
        pass  # don't worry this catches ConfigBlock too it seems
    else:
        config_file = val
        try:
            with open(config_file, "r") as f:
                val = json.load(f)
        except IOError:  # don't require config file
            _log.debug("Config file {} not found (this is okay)".format(config_file))
            return
    cfg.set_value(val)
    if config_file is not None:
        _log.debug("Read config {}".format(config_file))
    reconfig(cfg)

def write_config(path, cfg=None):
    if cfg is None:
        cfg = _new_idaes_config_block()
    with open(path, 'w') as f:
        json.dump(cfg.value(), f, cls=ConfigBlockJSONEncoder, indent=4)


def reconfig(cfg):
    logging.config.dictConfig(cfg.logging.value())
    setup_environment(bin_directory, cfg.use_idaes_solvers)


def create_dir(d):
    """Create a directory if it doesn't exist.

    Args:
        d(str): directory path to create

    Retruns:
        None
    """
    if os.path.exists(d):
        return
    else:
        os.mkdir(d)


def get_data_directory():
    """Return the standard data directory for idaes, based on the OS."""
    try:
        if os.name == 'nt':  # Windows
            data_directory = os.path.join(os.environ['LOCALAPPDATA'], "idaes")
        else:  # any other OS
            data_directory = os.path.join(os.environ['HOME'], ".idaes")
    except AttributeError:
        data_directory = None
    # Standard location for executable binaries.
    if data_directory is not None:
        bin_directory = os.path.join(data_directory, "bin")
    else:
        bin_directory = None
    # Standard location for testing files
    if data_directory is not None:
        testing_directory = os.path.join(data_directory, "testing")
    else:
        testing_directory = None

    return data_directory, bin_directory, testing_directory


def setup_environment(bin_directory, use_idaes_solvers):
    """
    Set environment variables for the IDAES session.

    Args:
        bin_directory: directory to find idaes libraries and executables
        use_idaes_solvers: If true look first in the idaes bin directory for
                           executables if false look last in the idaes bin
                           directory for executables.

    Returns:
        None
    """
    oe = orig_environ
    if use_idaes_solvers:
        os.environ['PATH'] = os.pathsep.join([bin_directory, oe.get('PATH', '')])
    else:
        os.environ['PATH'] = os.pathsep.join([oe.get('PATH', ''), bin_directory])
    if os.name != 'nt':  # If not Windwos set lib search path, Windows uses PATH
        os.environ['LD_LIBRARY_PATH'] = os.pathsep.join(
            [oe.get('LD_LIBRARY_PATH', ''), bin_directory])
        # This is for OSX, but won't hurt other UNIX
        os.environ['DYLD_LIBRARY_PATH'] = os.pathsep.join(
            [oe.get('DYLD_LIBRARY_PATH', ''), bin_directory])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
from idaes.surrogate import alamopy
from idaes.surrogate.pysmo import polynomial_regression, radial_basis_function, kriging
from idaes.surrogate.my_surrogate_base import Surrogate
from idaes.surrogate.my_surrogate_base import Metrics

from sympy.parsing.sympy_parser import parse_expr
from sympy import sympify
from sympy import symbols, var
import sympy
import time
import os.path, pickle

from pyomo.core.expr.sympy_tools import PyomoSympyBimap, sympy_available, Sympy2PyomoVisitor, sympy2pyomo_expression
import pyomo.environ as pyo
import numpy as np

try:
    import ujson as json
except:
    import json

from pyomo.common.config import ConfigBlock, ConfigValue, ConfigList


class GeneralSurrogate(Surrogate):
    CONFIG = Surrogate.CONFIG()

    CONFIG.declare('pyomo_vars', ConfigValue(default=None, domain=list))
    CONFIG.declare('linear', ConfigValue(default=True, domain=bool))
    CONFIG.declare('maximum_polynomial_order', ConfigValue(default=None, domain=int))
    CONFIG.declare('multinomials', ConfigValue(default=True, domain=bool))
    CONFIG.declare('logexp', ConfigValue(default=None, domain=bool))
    CONFIG.declare('sincos', ConfigValue(default=None, domain=bool))
    CONFIG.declare('ratio', ConfigValue(default=False, domain=bool))
    CONFIG.declare('alamo_modeler', ConfigValue(default=1, domain=int))
    CONFIG.declare('convpen', ConfigValue(default=0, domain=int))
    CONFIG.declare('basis_function', ConfigValue(default=None, domain=str))
    CONFIG.declare('regularization', ConfigValue(default=None, domain=bool))
    CONFIG.declare('metric', ConfigValue(default=Metrics.SSE, domain=str))
    CONFIG.declare('additional_features_list', ConfigValue(default=None, domain=list))
    CONFIG.declare('overwrite', ConfigValue(default=True, domain=bool))
    CONFIG.declare('fname', ConfigValue(default=None, domain=str))

    CONFIG.declare('alamopy', ConfigValue(default=True, domain=bool))
    CONFIG.declare('alamopy_rbf', ConfigValue(default=False, domain=bool))
    CONFIG.declare('pysmo_kriging', ConfigValue(default=False, domain=bool))
    CONFIG.declare('pysmo_rbf', ConfigValue(default=False, domain=bool))
    CONFIG.declare('pysmo_polyregression', ConfigValue(default=True, domain=bool))


    def __init__(self, **settings):
        super().__init__(**settings)

        self._pysmo_krg_settings = {}
        self._pysmo_rbf_settings = {}
        self._pysmo_pr_settings = {}
        self._alamo_settings = {}

        self._models = []

    def build_model(self):
        super().build_model()

        self.parse_config()
        # models = []

        modeler_krig = Pysmo_kriging(**self._pysmo_krg_settings)
        modeler_krig.regressed_data(self._rdata_in, self._rdata_out)
        if self.config['pysmo_kriging']:
            modeler_krig.build_model()
            self._models.append(modeler_krig)

        modeler_rbf = Pysmo_rbf(**self._pysmo_rbf_settings)
        modeler_rbf.regressed_data(self._rdata_in, self._rdata_out)
        if self.config['pysmo_rbf']:
            modeler_rbf.build_model()
            self._models.append(modeler_rbf)

        modeler_pr = Pysmo_polyregression(**self._pysmo_pr_settings)
        modeler_pr.regressed_data(self._rdata_in, self._rdata_out)
        if self.config['pysmo_polyregression']:
            modeler_pr.build_model()
            self._models.append(modeler_pr)

        modeler_alamo = Alamopy(**self._alamo_settings)
        modeler_alamo.regressed_data(self._rdata_in, self._rdata_out)
        if self.config['alamopy']:
            has_alamo_flag = alamopy.multos.has_alamo()
            if has_alamo_flag:
                modeler_alamo.build_model()
                self._models.append(modeler_alamo)
            else:
                print("ALAMO is not installed, and wasn't used for the General Surrogate.")

        _alamo_settings_rbf = self._alamo_settings
        if self.config['alamopy_rbf']:
            _alamo_settings_rbf['grbfcns'] = self.config['alamopy_rbf']
            modeler_alamo_rbf = Alamopy(**_alamo_settings_rbf)
            modeler_alamo_rbf.regressed_data(self._rdata_in, self._rdata_out)
            if has_alamo_flag:
                modeler_alamo_rbf.build_model()
                self._models.append(modeler_alamo_rbf)

        best_metric = 10e16
        best_config = {}
        for m in self._models:
            metric = m.get_results()[self.config['metric']]
            # print(m, m.get_results()['SSE'], best_metric, metric)
            if best_metric > float(metric):
                best_metric = metric
                best_config = m.config
                self._results = m.get_results()
                self._model = m.get_model()

        if not self.config['pysmo_rbf']:
            self._models.append(modeler_rbf)

        if not self.config['pysmo_kriging']:
            self._models.append(modeler_krig)

        self.pkl_info['Run settings'] = best_config
        self.pkl_info['Results'] = self._results
        self.pkl_info['Expression'] = self._model

    def parse_config(self):

        self._pysmo_krg_settings = {'numerical_gradients': True,
                                    'regularization': self.config['regularization'],
                                    'pyomo_vars': self.config['pyomo_vars'],
                                    'overwrite': self.config['overwrite']}

        if self.config['basis_function'] is None:
            self.config['basis_function'] = 'gaussian'

        self._pysmo_rbf_settings = {'basis_function': self.config['basis_function'],
                                    'regularization': self.config['regularization'],
                                    'pyomo_vars': self.config['pyomo_vars'],
                                    'overwrite': self.config['overwrite']}

        self._pysmo_pr_settings = {'maximum_polynomial_order': self.config['maximum_polynomial_order'],
                                   'multinomials': self.config['multinomials'],
                                   'pyomo_vars': self.config['pyomo_vars'],
                                   'training_split': 0.9,
                                   'number_of_crossvalidations': 5,
                                   'overwrite': self.config['overwrite']}

        if self.config['linear']:
            self._alamo_settings['linfcns'] = 1
        if self.config['logexp']:
            self._alamo_settings['logfcns'] = 1
            self._alamo_settings['expfcns'] = 1
        if self.config['sincos']:
            self._alamo_settings['sinfcns'] = 1
            self._alamo_settings['cosfcns'] = 1

        max_order = self.config['maximum_polynomial_order']
        pow_list = np.arange(0.5, max_order + 1, 0.5)
        if max_order >= 1:
            self._alamo_settings['monomialpower'] = pow_list
        if self.config['multinomials']:

            self._alamo_settings['multi2power'] = np.arange(0.5, max_order / 2 + 1, 0.5)
            self._alamo_settings['multi3power'] = np.arange(0.5, max_order / 2 + 1, 0.5)
            if self.config['ratio']:
                self._alamo_settings['ratiopower'] = pow_list
        self._alamo_settings['expandoutput'] = True
        self._alamo_settings['modeler'] = self.config['alamo_modeler']
        self._alamo_settings['convpen'] = self.config['convpen']

    def get_models(self):
        return self._models


class Alamopy(Surrogate):
    CONFIG = Surrogate.CONFIG()

    CONFIG.declare('xlabels', ConfigValue(default=None, domain=list))
    CONFIG.declare('zlabels', ConfigValue(default=None, domain=list))
    CONFIG.declare('xval', ConfigValue(default=None, domain=list))
    CONFIG.declare('zval', ConfigValue(default=None, domain=list))
    CONFIG.declare('xmin', ConfigValue(default=None, domain=list))
    CONFIG.declare('xmax', ConfigValue(default=None, domain=list))
    CONFIG.declare('modeler', ConfigValue(default=1, domain=int))
    CONFIG.declare('linfcns', ConfigValue(default=1, domain=int))
    CONFIG.declare('expfcns', ConfigValue(default=0, domain=int))
    CONFIG.declare('logfcns', ConfigValue(default=0, domain=int))
    CONFIG.declare('sinfcns', ConfigValue(default=0, domain=int))
    CONFIG.declare('cosfcns', ConfigValue(default=0, domain=int))
    CONFIG.declare('monomialpower', ConfigValue(default=None, domain=list))
    CONFIG.declare('multi2power', ConfigValue(default=None, domain=list))
    CONFIG.declare('multi3power', ConfigValue(default=None, domain=list))
    CONFIG.declare('ratiopower', ConfigValue(default=None, domain=list))
    CONFIG.declare('grbfcns', ConfigValue(default=0, domain=int))
    CONFIG.declare('convpen', ConfigValue(default=None, domain=float))
    CONFIG.declare('screener', ConfigValue(default=None, domain=int))  # not sure
    CONFIG.declare('almname', ConfigValue(default=None, domain=str))
    CONFIG.declare('showalm', ConfigValue(default=False, domain=bool))
    CONFIG.declare('savescratch', ConfigValue(default=None, domain=str))
    CONFIG.declare('savetrace', ConfigValue(default=None, domain=str))
    CONFIG.declare('expandoutput', ConfigValue(default=True, domain=bool))
    CONFIG.declare('almopt', ConfigValue(default=None, domain=str))
    CONFIG.declare('loo', ConfigValue(default=False, domain=bool))
    CONFIG.declare('lmo', ConfigValue(default=False, domain=bool))  # Check
    CONFIG.declare('maxiter', ConfigValue(default=None, domain=int))
    CONFIG.declare('simulator', ConfigValue(default=None))

    def __init__(self, **settings):
        super().__init__(**settings)
        self.alamopy_results = None

    def build_model(self):
        super().build_model()
        if self._vdata_in is not None and self._vdata_out is not None:
            self.alamopy_results = alamopy.alamo(self._rdata_in, self._rdata_out, xval=self._vdata_in,
                                                 zval=self._vdata_out, **self.config)
        else:
            self.alamopy_results = alamopy.alamo(self._rdata_in, self._rdata_out, **self.config)

        self._res = self.alamopy_results
        self._model = self.alamopy_results['model']

        self.handle_results(self._res)
        self.pkl_info['Run settings'] = self.config.value()
        self.pkl_info['Results'] = self._results
        self.pkl_info['Expression'] = self._model

    def handle_results(self, res):
        # sympy to pyomo converter

        self._results = {}
        if type(self.alamopy_results[Metrics.R2]) is dict:
            label = self.alamopy_results['zlabels'][0]
            self._results[Metrics.RMSE] = self.alamopy_results['rmse'][label]
            self._results[Metrics.SSE] = self.alamopy_results['ssr'][label]
            self._results[Metrics.Time] = self.alamopy_results['totaltime'][label]
            self._results[Metrics.MSE] = float(self.alamopy_results['rmse'][label]) ** 2
            self._results[Metrics.Order] = self.alamopy_results['nbas'][label]
            self._results[Metrics.R2] = self.alamopy_results['R2'][label]
        else:
            self._results[Metrics.RMSE] = self.alamopy_results['rmse']
            self._results[Metrics.SSE] = self.alamopy_results['ssr']
            self._results[Metrics.Time] = self.alamopy_results['totaltime']
            self._results[Metrics.MSE] = float(self.alamopy_results['rmse']) ** 2
            self._results[Metrics.Order] = self.alamopy_results['nbas']
            self._results[Metrics.R2] = self.alamopy_results['R2']

        # Generate pyomo expression
        m = pyo.ConcreteModel()
        m.x = pyo.Var(range(len(self._rdata_in)))

        obj_map = PyomoSympyBimap()
        obj_map.sympy2pyomo = {}
        sympy_locals = {}
        i = 1
        for label in res['xlabels']:
            sympy_locals[label] = sympy.Symbol(label)
            sympy_obj = sympy.Symbol(label)
            obj_map.sympy2pyomo[sympy_obj] = m.x[i]
            i += 1

        model_string = ""
        if type(res['model']) is dict:
            key = list(res['model'].keys())[0]
            model_string = res['model'][key].split('=')[1]
        else:
            model_string = res['model'].split('=')[1]
        model_symp = parse_expr(model_string.replace("^", "**"), local_dict=sympy_locals)
        model_pyomo = sympy2pyomo_expression(model_symp, obj_map)
        self._model = model_pyomo

    def run_test_suite(self, return_all=False, allow_grb=True):

        surrogate_modelers = []
        # all
        alamo_settings1 = {'logfcns': 1,
                           'expfcns': 1,
                           'monomialpower': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'multi2power': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'multi3power': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'ratiopower': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'sinfcns': 1,
                           'cosfcns': 1
                           }
        surrogate_modelers.append(Alamopy(**alamo_settings1))

        # all0
        alamo_settings2 = {'logfcns': 1,
                           'expfcns': 1,
                           'monomialpower': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'multi2power': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'multi3power': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           'ratiopower': (0.5, 1, 2, 3, -0.5, -1, -2, -3),
                           }
        surrogate_modelers.append(Alamopy(**alamo_settings2))

        # all1
        if allow_grb:
            alamo_settings3 = {'monomialpower': (0.1, 0.2, 0.4, 0.6, 0.8, 1, 1.2, 1.3, 1.4, 1.6, 1.8, 2, 2.5,
                                                 3, 3.5, 4, -0.1, -0.2, -0.4, -0.6, -0.8, -1, -1.2, -1.3, -1.4,
                                                 -1.6, -1.8, -2, -2.5, -3, -3.5, -4),
                               'grbfcns': 1
                               }
            surrogate_modelers.append(Alamopy(**alamo_settings3))

        # linear
        alamo_settings4 = {}
        surrogate_modelers.append(Alamopy(**alamo_settings4))

        # logexp
        alamo_settings5 = {'logfcns': 1,
                           'expfcns': 1
                           }
        surrogate_modelers.append(Alamopy(**alamo_settings5))

        # mono
        alamo_settings6 = {'monomialpower': (0.1, 0.2, 0.4, 0.6, 0.8, 1, 1.2, 1.3, 1.4, 1.6, 1.8, 2)}
        surrogate_modelers.append(Alamopy(**alamo_settings6))

        # mono2
        alamo_settings7 = {'monomialpower': (0.1, 0.2, 0.4, 0.6, 0.8, 1, 1.2, 1.3, 1.4, 1.6, 1.8, 2, 2.5,
                                             3, 3.5, 4, -0.1, -0.2, -0.4, -0.6, -0.8, -1, -1.2, -1.3, -1.4,
                                             -1.6, -1.8, -2, -2.5, -3, -3.5, -4)}
        surrogate_modelers.append(Alamopy(**alamo_settings7))

        # multi2
        alamo_settings8 = {'multi2power': (0.5, 1, 2, 3, -0.5, -1, -2, -3)}
        surrogate_modelers.append(Alamopy(**alamo_settings8))

        # multi3
        alamo_settings9 = {'multi3power': (0.5, 1, 2, 3, -0.5, -1, -2, -3)}
        surrogate_modelers.append(Alamopy(**alamo_settings9))

        # olr
        alamo_settings10 = {'modeler': 6, 'convpen': 0}
        surrogate_modelers.append(Alamopy(**alamo_settings10))

        # pce3
        alamo_settings11 = {'monomialpower': (1, 2, 3),
                            'multi2power': (1, 2)}
        surrogate_modelers.append(Alamopy(**alamo_settings11))

        # quad
        alamo_settings12 = {'monomialpower': (1, 2),
                            'multi2power': (1,)}
        surrogate_modelers.append(Alamopy(**alamo_settings12))

        # ratio
        alamo_settings13 = {'ratiopower': (0.5, 1, 2, 3, -0.5, -1, -2, -3)}
        surrogate_modelers.append(Alamopy(**alamo_settings13))

        # ratio2
        alamo_settings14 = {'ratiopower': (0.1, 0.2, 0.4, 0.6, 0.8, 1, 1.2,
                                           1.3, 1.4, 1.6, 1.8, 2, 2.5, 3, 3.5,
                                           4, -0.1, -0.2, -0.4, -0.6, -0.8, -1,
                                           -1.2, -1.3, -1.4, -1.6, -1.8, -2,
                                           -2.5, -3, -3.5, -4)}
        surrogate_modelers.append(Alamopy(**alamo_settings14))

        # rbf
        if (allow_grb):
            alamo_settings15 = {'grbfcns': 1}
            surrogate_modelers.append(Alamopy(**alamo_settings15))

        # sincos
        alamo_settings16 = {'sinfcns': 1,
                            'cosfcns': 1}
        surrogate_modelers.append(Alamopy(**alamo_settings16))

        best_model = None
        all_metrics = []
        for m in surrogate_modelers:
            m.regressed_data(self._rdata_in, self._rdata_out)
            m.build_model()
            metrics = m.get_results()
            if best_model is None or best_model.get_results()[Metrics.MSE] > metrics[Metrics.MSE]:
                best_model = m

            all_metrics.append(m)

        if return_all:
            return all_metrics

        return best_model


class Pysmo_rbf(Surrogate):
    CONFIG = Surrogate.CONFIG()

    CONFIG.declare('basis_function', ConfigValue(default=None, domain=str))
    CONFIG.declare('solution_method', ConfigValue(default=None, domain=str))
    CONFIG.declare('regularization', ConfigValue(default=None, domain=bool))
    CONFIG.declare('pyomo_vars', ConfigValue(default=None, domain=list))
    CONFIG.declare('overwrite', ConfigValue(default=True, domain=bool))
    CONFIG.declare('fname', ConfigValue(default=None, domain=str))

    def __init__(self, **settings):
        super().__init__(**settings)
        self.pysmo_rbf_results = None

    def build_model(self):
        super().build_model()
        start_time = time.time()
        training_data = np.concatenate((self._rdata_in, self._rdata_out.reshape(self._rdata_out.size, 1)), axis=1)
        self.pyomo_vars = dict(
            (k, self.config[k]) for k in ['pyomo_vars'] if k in self.config)  # Extract variable list
        self.rbf_setup = {k: self.config[k] for k in set(self.config) - set(self.pyomo_vars)}
        prob_init = radial_basis_function.RadialBasisFunctions(training_data, **self.rbf_setup)
        feature_vec = prob_init.get_feature_vector()
        self.pysmo_rbf_results = prob_init.training()
        self._results[Metrics.Time] = time.time() - start_time

        self.handle_results(feature_vec)

        self.pkl_info['Run settings'] = self.config
        self.pkl_info['Results'] = self._results
        self.pkl_info['pysmo result object'] = self.pysmo_rbf_results
        self.pkl_info['Expression'] = self._model
        self.pkl_info['Class initialization'] = prob_init

    def handle_results(self, feature_vec):
        self._results[Metrics.RMSE] = self.pysmo_rbf_results.rmse
        self._results[Metrics.SSE] = self.pysmo_rbf_results.rmse ** 2
        self._results[Metrics.R2] = self.pysmo_rbf_results.R2
        # Generate Pyomo expression
        if self.pyomo_vars:
            self._model = self.pysmo_rbf_results.generate_expression(self.pyomo_vars['pyomo_vars'])
        else:
            list_vars = []
            for i in feature_vec.keys():
                list_vars.append(feature_vec[i])
            # PYLINT-TODO-FIX the method "rbf_generate_expression" does not exist for this class
            # maybe the name of the method should be "generate_expression" like in the self.pyomo_vars case?
            # pylint: disable=no-member
            self._model = self.pysmo_rbf_results.rbf_generate_expression(list_vars)


class Pysmo_kriging(Surrogate):
    CONFIG = Surrogate.CONFIG()

    CONFIG.declare('numerical_gradients', ConfigValue(default=None, domain=bool))
    CONFIG.declare('regularization', ConfigValue(default=None, domain=bool))
    CONFIG.declare('pyomo_vars', ConfigValue(default=None, domain=list))
    CONFIG.declare('overwrite', ConfigValue(default=True, domain=bool))
    CONFIG.declare('fname', ConfigValue(default=None, domain=str))

    def __init__(self, **settings):
        super().__init__(**settings)
        self.pysmo_kriging_results = None

    def build_model(self):
        super().build_model()
        start_time = time.time()
        training_data = np.concatenate((self._rdata_in, self._rdata_out.reshape(self._rdata_out.size, 1)), axis=1)
        self.pyomo_vars = dict(
            (k, self.config[k]) for k in ['pyomo_vars'] if k in self.config)  # Extreact variable list
        self.krg_setup = {k: self.config[k] for k in set(self.config) - set(self.pyomo_vars)}
        prob_init = kriging.KrigingModel(training_data, **self.krg_setup)
        feature_vec = prob_init.get_feature_vector()
        self.pysmo_kriging_results = prob_init.training()
        self._results[Metrics.Time] = time.time() - start_time

        self.handle_results(feature_vec)

        self.pkl_info['pysmo result object'] = self.pysmo_kriging_results
        self.pkl_info['Class initialization'] = prob_init
        self.pkl_info['Run settings'] = self.config
        self.pkl_info['Results'] = self._results
        self.pkl_info['Expression'] = self._model

    def handle_results(self, feature_vec):
        self._results[Metrics.RMSE] = self.pysmo_kriging_results.training_rmse
        self._results[Metrics.SSE] = self.pysmo_kriging_results.training_rmse ** 2
        self._results[Metrics.R2] = self.pysmo_kriging_results.training_R2

        # Generate Pyomo expression
        list_vars = []
        for i in feature_vec.keys():
            list_vars.append(feature_vec[i])
        if self.pyomo_vars:
            self._model = self.pysmo_kriging_results.generate_expression(self.pyomo_vars['pyomo_vars'])
        else:
            list_vars = []
            for i in feature_vec.keys():
                list_vars.append(feature_vec[i])
            # PYLINT-TODO-FIX the method "kriging_generate_expression" does not exist for this class
            # maybe the name of the method should be "generate_expression" like in the self.pyomo_vars case?
            # pylint: disable=no-member
            self._model = self.pysmo_kriging_results.kriging_generate_expression(list_vars)


class Pysmo_polyregression(Surrogate):
    CONFIG = Surrogate.CONFIG()

    CONFIG.declare('number_of_crossvalidations', ConfigValue(default=None, domain=int))
    CONFIG.declare('maximum_polynomial_order', ConfigValue(default=None, domain=int))
    CONFIG.declare('training_split', ConfigValue(default=None, domain=float))
    CONFIG.declare('solution_method', ConfigValue(default=None, domain=str))
    CONFIG.declare('multinomials', ConfigValue(default=None, domain=bool))
    CONFIG.declare('additional_features_list', ConfigValue(default=None, domain=list))
    CONFIG.declare('pyomo_vars', ConfigValue(default=None, domain=list))
    CONFIG.declare('overwrite', ConfigValue(default=True, domain=bool))
    CONFIG.declare('fname', ConfigValue(default=None, domain=str))

    def __init__(self, **settings):
        super().__init__(**settings)
        self.pysmo_kriging_results = None

    def build_model(self):
        super().build_model()
        start_time = time.time()
        training_data = np.concatenate((self._rdata_in, self._rdata_out.reshape(self._rdata_out.size, 1)), axis=1)
        self.pyomo_vars = dict((k, self.config[k]) for k in ['pyomo_vars'] if k in self.config)  # Extract variable list
        self.add_terms = dict((k, self.config[k]) for k in ['additional_features_list'] if k in self.config)
        self.polyreg_setup = {k: self.config[k] for k in set(self.config) - set(self.pyomo_vars) - set(self.add_terms)}
        prob_init = polynomial_regression.PolynomialRegression(training_data, training_data, **self.polyreg_setup)
        feature_vec = prob_init.get_feature_vector()

        if self.config['additional_features_list'] is not None:
            try:
                term_lst = []
                res = [sub.replace('ft', 'feature_vec') for sub in self.add_terms['additional_features_list']]
                for i in res:
                    term_lst.append(eval(i))
                prob_init.set_additional_terms(term_lst)
            except:
                raise ValueError('The additional terms could not be evaluated.')

        self.pysmo_polyregression_results = prob_init.training()
        self._results[Metrics.Time] = time.time() - start_time

        self.handle_results(feature_vec)

        self.pkl_info['pysmo result object'] = self.pysmo_polyregression_results
        self.pkl_info['Class initialization'] = prob_init
        self.pkl_info['Run settings'] = self.config
        self.pkl_info['Results'] = self._results
        self.pkl_info['Expression'] = self._model

    def handle_results(self, feature_vec):
        self._results[Metrics.RMSE] = np.sqrt(self.pysmo_polyregression_results.errors['MSE'])
        self._results[Metrics.SSE] = self.pysmo_polyregression_results.errors['MSE']
        self._results[Metrics.R2] = self.pysmo_polyregression_results.errors['R2']

        # Generate Pyomo expression
        list_vars = []
        for i in feature_vec.keys():
            list_vars.append(feature_vec[i])
        if self.pyomo_vars:
            self._model = self.pysmo_polyregression_results.generate_expression(self.pyomo_vars['pyomo_vars'])
        else:
            list_vars = []
            for i in feature_vec.keys():
                list_vars.append(feature_vec[i])
            self._model = self.pysmo_polyregression_results.generate_expression(list_vars)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           units)
from pyomo.util.check_units import (assert_units_consistent,
                                    assert_units_equivalent)

from idaes.core import FlowsheetBlock, EnergyBalanceType, MomentumBalanceType
from idaes.generic_models.unit_models.gibbs_reactor import GibbsReactor
from idaes.generic_models.properties.activity_coeff_models.methane_combustion_ideal \
    import MethaneParameterBlock as MethaneCombustionParameterBlock
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (PhysicalParameterTestBlock,
                                     initialization_tester)
from idaes.core.util import get_solver
from idaes.core.util.exceptions import ConfigurationError


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = GibbsReactor(default={"property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.inert_species == []

    assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
    assert len(m.fs.unit.gibbs_minimization) == 4

    assert not hasattr(m.fs.unit, "inert_species_balance")


@pytest.mark.unit
def test_inerts():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = GibbsReactor(default={"property_package": m.fs.properties,
                                      "inert_species": ["c1"]})

    assert isinstance(m.fs.unit.inert_species_balance, Constraint)
    assert len(m.fs.unit.inert_species_balance) == 2
    assert m.fs.unit.inert_species_balance[0, "p1", "c1"] != Constraint.Skip
    assert m.fs.unit.inert_species_balance[0, "p2", "c1"] != Constraint.Skip

    assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
    assert len(m.fs.unit.gibbs_minimization) == 2


@pytest.mark.unit
def test_inerts_dependent_w_multi_phase():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    # Change elemental composition to introduce dependency
    m.fs.properties.element_comp = {"c1": {"H": 0, "He": 0, "Li": 3},
                                    "c2": {"H": 4, "He": 5, "Li": 0}}

    m.fs.unit = GibbsReactor(default={"property_package": m.fs.properties,
                                      "inert_species": ["c1"]})

    assert isinstance(m.fs.unit.inert_species_balance, Constraint)
    assert len(m.fs.unit.inert_species_balance) == 2
    assert m.fs.unit.inert_species_balance[0, "p1", "c1"] != Constraint.Skip
    assert m.fs.unit.inert_species_balance[0, "p2", "c1"] != Constraint.Skip

    assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
    assert len(m.fs.unit.gibbs_minimization) == 2


@pytest.mark.unit
def test_inerts_dependent_w_single_phase():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    # Set phase list to only have 1 phase
    m.fs.properties.phase_list = ["p1"]
    # Change elemental composition to introduce dependency
    m.fs.properties.element_comp = {"c1": {"H": 0, "He": 0, "Li": 3},
                                    "c2": {"H": 4, "He": 5, "Li": 0}}

    m.fs.unit = GibbsReactor(default={"property_package": m.fs.properties,
                                      "inert_species": ["c1"]})

    assert isinstance(m.fs.unit.inert_species_balance, Constraint)
    assert len(m.fs.unit.inert_species_balance) == 0
    assert (0, "p1", "c1") not in m.fs.unit.inert_species_balance

    assert isinstance(m.fs.unit.gibbs_minimization, Constraint)
    assert len(m.fs.unit.gibbs_minimization) == 1


@pytest.mark.unit
def test_invalid_inert():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    with pytest.raises(ConfigurationError,
                       match="fs.unit invalid component in inert_species "
                       "argument. foo is not in the property package "
                       "component list."):
        m.fs.unit = GibbsReactor(default={"property_package": m.fs.properties,
                                          "inert_species": ["foo"]})


# -----------------------------------------------------------------------------
class TestMethane(object):
    @pytest.fixture(scope="class")
    def methane(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = MethaneCombustionParameterBlock()

        m.fs.unit = GibbsReactor(default={
                "property_package": m.fs.properties,
                "has_heat_transfer": True,
                "has_pressure_change": True})

        m.fs.unit.inlet.flow_mol[0].fix(230.0)
        m.fs.unit.inlet.mole_frac_comp[0, "H2"].fix(0.0435)
        m.fs.unit.inlet.mole_frac_comp[0, "N2"].fix(0.6522)
        m.fs.unit.inlet.mole_frac_comp[0, "O2"].fix(0.1739)
        m.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "CH4"].fix(0.1304)
        m.fs.unit.inlet.mole_frac_comp[0, "CO"].fix(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(1e-5)
        m.fs.unit.inlet.mole_frac_comp[0, "NH3"].fix(1e-5)
        m.fs.unit.inlet.temperature[0].fix(1500.0)
        m.fs.unit.inlet.pressure[0].fix(101325.0)

        m.fs.unit.outlet.temperature[0].fix(2844.38)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, methane):
        assert hasattr(methane.fs.unit, "inlet")
        assert len(methane.fs.unit.inlet.vars) == 4
        assert hasattr(methane.fs.unit.inlet, "flow_mol")
        assert hasattr(methane.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(methane.fs.unit.inlet, "temperature")
        assert hasattr(methane.fs.unit.inlet, "pressure")

        assert hasattr(methane.fs.unit, "outlet")
        assert len(methane.fs.unit.outlet.vars) == 4
        assert hasattr(methane.fs.unit.outlet, "flow_mol")
        assert hasattr(methane.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(methane.fs.unit.outlet, "temperature")
        assert hasattr(methane.fs.unit.outlet, "pressure")

        assert hasattr(methane.fs.unit, "gibbs_minimization")
        assert hasattr(methane.fs.unit, "heat_duty")
        assert hasattr(methane.fs.unit, "deltaP")

        assert number_variables(methane) == 80
        assert number_total_constraints(methane) == 67
        assert number_unused_variables(methane) == 0

    @pytest.mark.component
    def test_units(self, methane):
        assert_units_consistent(methane)
        assert_units_equivalent(methane.fs.unit.heat_duty[0], units.W)
        assert_units_equivalent(methane.fs.unit.deltaP[0], units.Pa)

    @pytest.mark.unit
    def test_dof(self, methane):
        assert degrees_of_freedom(methane) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_temperature(self, methane):
        initialization_tester(methane,
                              optarg={'tol': 1e-6},
                              state_args={'temperature': 2844.38,
                                          'pressure': 101325.0,
                                          'flow_mol': 251.05,
                                          'mole_frac_comp': {'CH4': 1e-5,
                                                             'CO': 0.0916,
                                                             'CO2': 0.0281,
                                                             'H2': 0.1155,
                                                             'H2O': 0.1633,
                                                             'N2': 0.5975,
                                                             'NH3': 1e-5,
                                                             'O2': 0.0067}})

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_temperature(self, methane):
        results = solver.solve(methane)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution_temperature(self, methane):
        assert (pytest.approx(250.06, abs=1e-2) ==
                value(methane.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(0.0, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CH4"]))
        assert (pytest.approx(0.0974, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO"]))
        assert (pytest.approx(0.0226, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO2"]))
        assert (pytest.approx(0.1030, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2"]))
        assert (pytest.approx(0.1769, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2O"]))
        assert (pytest.approx(0.5999, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "N2"]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "NH3"]))
        assert (pytest.approx(0.0002, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "O2"]))
        assert (pytest.approx(-7454077, abs=1e2) ==
                value(methane.fs.unit.heat_duty[0]))
        assert (pytest.approx(101325.0, abs=1e-2) ==
                value(methane.fs.unit.outlet.pressure[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation_temperature(self, methane):
        assert abs(value(
                methane.fs.unit.inlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_in[0]
                    .enth_mol_phase["Vap"] -
                methane.fs.unit.outlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_out[0]
                    .enth_mol_phase["Vap"] +
                methane.fs.unit.heat_duty[0])) <= 1e-6

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_duty(self, methane):
        methane.fs.unit.outlet.temperature[0].unfix()
        methane.fs.unit.heat_duty.fix(-7454077)
        assert degrees_of_freedom(methane) == 0

        orig_fixed_vars = fixed_variables_set(methane)
        orig_act_consts = activated_constraints_set(methane)

        methane.fs.unit.initialize(
                          optarg={'tol': 1e-6},
                          state_args={'temperature': 2844.38,
                                      'pressure': 101325.0,
                                      'flow_mol': 251.05,
                                      'mole_frac_comp': {'CH4': 1e-5,
                                                         'CO': 0.0916,
                                                         'CO2': 0.0281,
                                                         'H2': 0.1155,
                                                         'H2O': 0.1633,
                                                         'N2': 0.5975,
                                                         'NH3': 1e-5,
                                                         'O2': 0.0067}})

        assert degrees_of_freedom(methane) == 0

        fin_fixed_vars = fixed_variables_set(methane)
        fin_act_consts = activated_constraints_set(methane)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_heat_duty(self, methane):
        solver.options["tol"] = 1e-9
        solver.options["nlp_scaling_method"] = "user-scaling"

        results = solver.solve(methane, tee=True)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution_duty(self, methane):
        assert (pytest.approx(250.06, abs=1e-1) ==
                value(methane.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(0.0, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CH4"]))
        assert (pytest.approx(0.103863, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO"]))
        assert (pytest.approx(0.016123, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO2"]))
        assert (pytest.approx(0.096080, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2"]))
        assert (pytest.approx(0.183897, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2O"]))
        assert (pytest.approx(0.600030, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "N2"]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "NH3"]))
        assert (pytest.approx(2.86950e-06, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "O2"]))
        assert (pytest.approx(-7454077, abs=1e2) ==
                value(methane.fs.unit.heat_duty[0]))
        assert (pytest.approx(101325.0, abs=1e-2) ==
                value(methane.fs.unit.outlet.pressure[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation_duty(self, methane):
        assert abs(value(
                methane.fs.unit.inlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_in[0]
                    .enth_mol_phase["Vap"] -
                methane.fs.unit.outlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_out[0]
                    .enth_mol_phase["Vap"] +
                methane.fs.unit.heat_duty[0])) <= 1e-4

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, methane):
        methane.fs.unit.report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Search through the code and index static information in the DMF.
"""
# stdlib
import glob
import importlib
import inspect
import logging
import os
import pprint
import re

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

_log = logging.getLogger(__name__)


class Walker(object):
    def walk(self, visitor):
        """Interface for walkers.

        Args:
            visitor (Visitor): Class whose `visit` method will be
                               called for each item.

        Returns:
            None
        """
        pass


class ModuleClassWalker(Walker):
    """Walk modules from a given root (e.g. 'idaes'), and visit
    all classes in those modules whose name matches a given pattern.

    Example usage::

        walker = ModuleClassWalker(from_pkg=idaes,
                                   class_expr='_PropertyParameter.*')

        walker.walk(PrintMetadataVisitor())  # see below

    """

    def __init__(
        self,
        from_path=None,
        from_pkg=None,
        class_expr=None,
        parent_class=None,
        suppress_warnings=False,
        exclude_testdirs=True,
        exclude_tests=True,
        exclude_init=True,
        exclude_setup=True,
        exclude_dirs=None,
    ):
        """Constructor. Create from either a path or package root, or both.
        If just one is given, the other is deduced. At least one must be
        given.

        Args:
            from_path (str): Path to start looking for modules. Overrides
                            'from_pkg'.
            from_pkg (module): Package from which to start looking for modules.
                            The path of the package will depend on the Python
                            environment and path.
            class_expr (str): Regular expression for the classes to find.
            parent_class (class): If given, filter for this parent class.
                                  Will match class_expr OR parent_class.
            suppress_warnings (bool): If True do not log any warning messages.
            exclude_testdirs (bool): If True (the default), exclude test dirs
            exclude_tests (bool): If True (the default), exclude test files
            exclude_init (bool): If True (the default), exclude __init___.py
            exclude_setup (bool): If True (the default), exclude setup.py
            exclude_dirs (list): List of directory (regex) patterns to exclude.
                                 These will be prefixed with a directory
                                 separator (e.g. '/') but otherwise can be
                                 any path expression.
        """
        if from_path:
            self._root = from_path
            if from_pkg:
                self._pkg = from_pkg.__name__
            else:
                self._pkg = str(os.path.basename(self._root))
        elif from_pkg:
            self._pkg = from_pkg.__name__
            try:
                self._root = from_pkg.__path__[0]
            except AttributeError:
                raise IOError(
                    "Root directory cannot be deduced from"
                    ' package "{}": no __path__ attribute.'.format(self._pkg)
                )
        else:
            raise ValueError(
                'Missing arguments: either "from_pkg" or ' '"from_path" must be given'
            )
        if not os.path.isdir(self._root):
            raise IOError('Root directory "{}"'.format(self._root))
        self._expr = re.compile(class_expr) if class_expr else None
        self._warn = not suppress_warnings
        self._parent = parent_class
        # build regular expression of things to exclude
        expr_list = []
        psep = os.path.sep
        if psep == "\\":
            psep = "\\\\"
        if exclude_testdirs:
            expr_list.append(r"{sl}tests?{sl}".format(sl=psep))
        if exclude_tests:
            expr_list.append(r"tests?_.*.py")
        if exclude_init:
            expr_list.append(r"__init__\.py")
        if exclude_setup:
            expr_list.append(r"setup\.py")
        if exclude_dirs:
            for ed in exclude_dirs:
                expr_list.append("{sl}{d}".format(sl=psep, d=ed))
        self._exclude_expr = re.compile(r"|".join(expr_list))
        self._history = []
        _log.debug("exclude expr={}".format(self._exclude_expr.pattern))

    def walk(self, visitor):
        modules = self._get_modules()
        self._visit_subclasses(modules, visitor.visit)

    def get_indexed_classes(self):
        return self._history

    def _get_modules(self):
        _log.debug("getting modules from root: {}".format(self._root))
        # change file paths at 'root' to module paths from 'pkgroot'
        n, module_list = len(self._root), []
        for f in self._python_files():
            module_path = os.path.splitext(f[n + 1 :])[0]
            module_path = module_path.replace(os.path.sep, ".")
            module_list.append(self._pkg + "." + module_path)
        return module_list

    def _python_files(self):
        q = [self._root]
        while q:
            curdir = q.pop()
            for f in glob.glob(os.path.join(curdir, "*.py")):
                if not self._exclude_expr.search(f):
                    yield f
            for d in os.listdir(curdir):
                path = os.path.join(curdir, d)
                if os.path.isdir(path):
                    q.append(path)

    def _visit_subclasses(self, modules, visit):
        for modname in modules:
            _log.debug("visit module: {}".format(modname))
            try:
                mod = importlib.import_module(modname)
            except Exception:
                if self._warn:
                    _log.warn(
                        "Error during import of module: {}. Ignoring.".format(modname)
                    )
                continue
            for item in dir(mod):
                x = getattr(mod, item)
                if inspect.isclass(x):
                    fullname = modname + "." + x.__name__
                    if not self._expr and not self._parent:
                        if visit(x):
                            self._history.append(fullname)
                    elif self._expr:
                        if self._expr.match(x.__name__):
                            if visit(x):
                                self._history.append(fullname)
                    elif self._parent and issubclass(x, self._parent):
                        if visit(x):
                            self._history.append(fullname)


class Visitor(object):
    """Interface for the 'visitor' class passed to Walker subclasses'
    `walk()` method.
    """

    def visit(self, obj):
        """Visit one object.

        Args:
            obj (object): Some object to operate on.

        Returns:
            True if visit succeeded, else False
        """
        pass


class PropertyMetadataVisitor(Visitor):
    """Visit something implementing :class:`HasPropertyClassMetadata` and
    pass that metadata, as a dict, to the `visit_metadata()` method,
    which should be implemented by the subclass.
    """

    def visit(self, obj):
        """Visit one object.

        Args:
            obj (idaes.core.property_base.HasPropertyClassMetadata): The object

        Returns:
            True if visit succeeded, else False
        """
        result, meta = True, None
        try:
            meta = obj.get_metadata()
        except (AttributeError, TypeError) as err:
            module = obj.__module__
            _log.debug(
                "Cannot get metadata for {}.{}: {}".format(module, obj.__name__, err)
            )
            result = False
        except NotImplementedError:
            module = obj.__module__
            #            if not module.startswith('idaes.core.'):
            _log.warn(
                '{} in module "{}" does not define its metadata'.format(
                    obj.__name__, module
                )
            )
            result = False
        if result:
            self.visit_metadata(obj, meta)
        return result

    def visit_metadata(self, obj, meta):
        """Do something with the metadata.

        Args:
            obj (object): Object from which metadata was pulled, for context.
            meta (idaes.core.property_base.PropertyClassMetadata): The metadata

        Returns:
            None
        """
        pass


class PrintPropertyMetadataVisitor(PropertyMetadataVisitor):
    def visit_metadata(self, obj, meta):
        """Print the module and class of the object, and
        then the metadata dict, to standard output.
        """
        meta_dict = {"properties": meta.properties, "default_units": meta.default_units}
        print("----")
        print("{}: class {}".format(obj.__module__, obj.__name__))
        print("  metadata: {}".format(pprint.pformat(meta_dict)))


# This is for testing


class _TestClass(object):
    pass


class _TestClass1(_TestClass):
    @classmethod
    def get_metadata(cls):
        return {"x": 1}


class _TestClass2(_TestClass):
    @classmethod
    def get_metadata(cls):
        return {"x": 2}


class _TestClass3(_TestClass):
    @classmethod
    def get_metadata(cls):
        return {"x": 3}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""IDAES Span-Wager CO2 properties

Dropped all critical enhancments and non-analytic terms ment to improve
accruacy near the critical point. These tend to cause singularities in the
equations, and it is assumend that we will try to avoid operating very close to
the critical point.

 References: (some of this is only used in the C++ part)
   Span, R., W. Wagner, W. (1996). "A New Equation of State for Carbon Dioxide
       Covering the Fluid Region from the Triple-Point Temperature to 1100 K at
       Pressures up to 800 MPa." J. Phys. Chem. Ref. Data, 25(6), 1509-1596.
   Akasaka, R. (2008). "A Reliable and Useful Method to Determine the
       Saturation State from Helmholtz Energy Equations of State." Journal of
       Thermal Science and Technology, 3(3), 442-451.
   Vesovic, V., W.A. Wakeham, G.A. Olchowy, J.V. Sengers, J.T.R. Watson, J.
       Millat, (1990). "The transport properties of carbon dioxide." J. Phys.
       Chem. Ref. Data, 19, 763-808.
   Fenghour, A., W.A. Wakeham, V. Vesovic, (1998). "The Viscosity of Carbon
       Dioxide." J. Phys. Chem. Ref. Data, 27, 31-44.
"""
__author__ = "John Eslick"

# Import Python libraries
import os

# Import Pyomo libraries
from pyomo.environ import (
    Expression,
    Param,
    RangeSet,
    Set,
    exp,
    sqrt,
    log,
    units as pyunits
)

# Import IDAES
from idaes.core import (
    declare_process_block_class,
    StateBlock,
    StateBlockData,
)
import idaes
import idaes.logger as idaeslog
from idaes.generic_models.properties.helmholtz.helmholtz import (
    _available,
    _htpx,
    HelmholtzParameterBlockData,
    HelmholtzStateBlockData,
    HelmholtzThermoExpressions,
    PhaseType,
    StateVars,
    _StateBlock,
)

# Logger
_log = idaeslog.getLogger(__name__)
_so = os.path.join(idaes.bin_directory, "swco2_external.so")


def swco2_available():
    """Make sure the compiled IAPWS-95 functions are available. Yes, in Windows
    the .so extention is still used.
    """
    return _available(_so)


def htpx(T=None, P=None, x=None):
    """
    Convenience function to calculate enthalpy from temperature and
    either pressure or vapor fraction. This function can be used for inlet
    streams and initialization where temperature is known instead of enthalpy.
    User must provided values for two of T, P, or x.

    Args:
        T: Temperature with units (between 200 and 3000 K)
        P: Pressure with units (between 1 and 1e9 Pa), None if saturated vapor
        x: Vapor fraction [mol vapor/mol total] (between 0 and 1), None if
           superheated or subcooled

    Returns:
        Total molar enthalpy [J/mol].
    """
    prop = SWCO2StateBlock(default={"parameters": SWCO2ParameterBlock()})
    return _htpx(T=T, P=P, x=x, prop=prop,
                 Tmin=200*pyunits.K, Tmax=3e3*pyunits.K,
                 Pmin=1e-4*pyunits.kPa, Pmax=1e6*pyunits.kPa)


@declare_process_block_class("SWCO2ParameterBlock")
class SWCO2ParameterBlockData(HelmholtzParameterBlockData):
    CONFIG = HelmholtzParameterBlockData.CONFIG()

    def build(self):
        self._set_parameters(
            library=_so,
            eos_tag="swco2",
            state_block_class=SWCO2StateBlock,
            component_list=Set(initialize=["CO2"]),
            phase_equilibrium_idx=Set(initialize=[1]),
            phase_equilibrium_list={1: ["CO2", ("Vap", "Liq")]},
            mw=Param(initialize=0.0440098,
                     doc="Molecular weight [kg/mol]",
                     units=pyunits.kg/pyunits.mol),
            temperature_crit=Param(
                initialize=304.1282,
                doc="Critical temperature [K]",
                units=pyunits.K),
            pressure_crit=Param(initialize=7.377e6,
                                doc="Critical pressure [Pa]",
                                units=pyunits.Pa),
            dens_mass_crit=Param(initialize=467.6,
                                 doc="Critical density [kg/m3]",
                                 units=pyunits.kg/pyunits.m**3),
            specific_gas_constant=Param(
                initialize=188.9241,
                doc="CO2 Specific Gas Constant [J/kg/K]",
                units=pyunits.J/pyunits.kg/pyunits.K),
            pressure_bounds=(0.1, 1e9),
            temperature_bounds=(210, 2500),
            enthalpy_bounds=(-2e4, 1e5),
        )

        super().build()

        # Thermal conductivity parameters.
        # Vesovic et al. (1990)
        self.tc_b = Param(
            RangeSet(0, 8),
            initialize={
                0: 0.4226159,
                1: 0.6280115,
                2: -0.5387661,
                3: 0.6735941,
                4: 0,
                5: 0,
                6: -0.4362677,
                7: 0.2255388,
            },
            doc="0 density limit thermal conductivity parameters",
        )

        self.tc_c = Param(
            RangeSet(1, 6),
            initialize={
                1: 2.387869e-2,
                2: 4.350794,
                3: -10.33404,
                4: 7.981590,
                5: -1.940558,
            },
            doc="0 density limit thermal conductivity parameters",
        )

        self.tc_d_1 = Param(
            initialize=2.447164e-5*self.dens_mass_crit,
            doc="Residual thermal conductivity parameter",
            units=pyunits.W/pyunits.K/pyunits.m)
        self.tc_d_2 = Param(
            initialize=8.705605e-8*self.dens_mass_crit**2,
            doc="Residual thermal conductivity parameter",
            units=pyunits.W/pyunits.K/pyunits.m)
        self.tc_d_3 = Param(
            initialize=-6.547950e-11*self.dens_mass_crit**3,
            doc="Residual thermal conductivity parameter",
            units=pyunits.W/pyunits.K/pyunits.m)
        self.tc_d_4 = Param(
            initialize=6.594919e-14*self.dens_mass_crit**4,
            doc="Residual thermal conductivity parameter",
            units=pyunits.W/pyunits.K/pyunits.m)

        # Viscosity parameters
        # "Fenghour et al. (1998) with critial enhancment Vesovic et al. (1990)
        self.visc_a = Param(
            RangeSet(0, 5),
            initialize={
                0: 0.235156,
                1: -0.491266,
                2: 5.211155e-2,
                3: 5.347906e-2,
                4: -1.537102e-2,
            },
            doc="0 density limit viscosity parameters",
        )

        # The indexing looks a little weird here, but it's from the source
        self.visc_d_1_1 = Param(
            initialize=0.4071119e-8*self.dens_mass_crit,
            doc="Residual viscosity parameter",
            units=pyunits.Pa*pyunits.s)
        self.visc_d_2_1 = Param(
            initialize=0.7198037e-10*self.dens_mass_crit**2,
            doc="Residual viscosity parameter",
            units=pyunits.Pa*pyunits.s)
        self.visc_d_6_4 = Param(
            initialize=0.2411697e-22*self.dens_mass_crit**6,
            doc="Residual viscosity parameter",
            units=pyunits.Pa*pyunits.s)
        self.visc_d_8_1 = Param(
            initialize=0.2971072e-28*self.dens_mass_crit**8,
            doc="Residual viscosity parameter",
            units=pyunits.Pa*pyunits.s)
        self.visc_d_8_2 = Param(
            initialize=-0.1627888e-28*self.dens_mass_crit**8,
            doc="Residual viscosity parameter",
            units=pyunits.Pa*pyunits.s)

        self.set_default_scaling("therm_cond_phase", 1e2, index="Liq")
        self.set_default_scaling("therm_cond_phase", 1e1, index="Vap")
        self.set_default_scaling("visc_d_phase", 1e5, index="Liq")
        self.set_default_scaling("visc_d_phase", 1e6, index="Vap")
        self.set_default_scaling("visc_k_phase", 1e5, index="Liq")
        self.set_default_scaling("visc_k_phase", 1e7, index="Vap")


@declare_process_block_class(
    "SWCO2StateBlock",
    block_class=_StateBlock,
    doc="""This is some placeholder doc.""",
)
class SWCO2StateBlockData(HelmholtzStateBlockData):
    """
    This is a property package for calculating thermophysical properties of
    water.
    """

    def build(self, *args):
        """
        Callable method for Block construction
        """
        super().build(*args)

        phlist = self.config.parameters.private_phase_list
        T = self.temperature
        rho = self.dens_mass_phase
        delta = self.dens_phase_red

        # Phase Thermal conductiviy
        def rule_tc(b, p):
            params = self.config.parameters
            b = params.tc_b
            c = params.tc_c
            cint_over_k = (
                1.0 + exp(-183.5*pyunits.K/T) *
                sum(c[i]*(T/100/pyunits.K)**(2-i) for i in c)
            )
            Ts = T/(251.196*pyunits.K)
            G = sum(b[i]/Ts**i for i in b)
            return (
                0.475598/1e3 * pyunits.W/pyunits.K/pyunits.m *
                   sqrt(T/pyunits.K)*(1 + 2.0/5.0*cint_over_k)/G +
                params.tc_d_1*delta[p] +
                params.tc_d_2*delta[p]**2 +
                params.tc_d_3*delta[p]**3 +
                params.tc_d_4*delta[p]**4)

        self.therm_cond_phase = Expression(
            phlist,
            rule=rule_tc,
            doc="Thermal conductivity [W/K/m]")

        # Phase dynamic viscosity
        def rule_mu(b, p):
            params = self.config.parameters
            a = params.visc_a
            Ts = T/(251.196*pyunits.K)
            return (
                1.00697/1e6*pyunits.Pa*pyunits.s*sqrt(T/pyunits.K) /
                    exp(sum(a[i]*log(Ts)**i for i in a)) +
                params.visc_d_1_1*delta[p] +
                params.visc_d_2_1*delta[p]**2 +
                params.visc_d_6_4*delta[p]**6/Ts**3 +
                params.visc_d_8_1*delta[p]**8 +
                params.visc_d_8_2*delta[p]**8/Ts)

        self.visc_d_phase = Expression(
            phlist,
            rule=rule_mu,
            doc="Viscosity (dynamic)"
        )

        # Phase kinimatic viscosity
        def rule_nu(b, p):
            return self.visc_d_phase[p] / self.dens_mass_phase[p]

        self.visc_k_phase = Expression(
            phlist, rule=rule_nu, doc="Kinematic viscosity"
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for CLC gas phase thermo state block; tests for construction and solve
Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           Var)

from idaes.core import FlowsheetBlock

from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core.util.testing import (get_default_solver,
                                     initialization_tester)

from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
    gas_phase_thermo import GasPhaseParameterBlock

# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
@pytest.fixture(scope="class")
def gas_prop():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    # gas properties and state inlet block
    m.fs.properties = GasPhaseParameterBlock()
    m.fs.unit = m.fs.properties.build_state_block(
        default={"parameters": m.fs.properties,
                 "defined_state": True})

    m.fs.unit.flow_mol.fix(1)
    m.fs.unit.temperature.fix(450)
    m.fs.unit.pressure.fix(1.60)
    m.fs.unit.mole_frac_comp["CO2"].fix(0.0004)
    m.fs.unit.mole_frac_comp["H2O"].fix(0.0093)
    m.fs.unit.mole_frac_comp["O2"].fix(0.2095)
    m.fs.unit.mole_frac_comp["N2"].fix(0.7808)

    return m


@pytest.mark.unit
def test_build_inlet_state_block(gas_prop):
    assert isinstance(gas_prop.fs.unit.mw, Var)
    assert isinstance(gas_prop.fs.unit.dens_mol, Var)
    assert isinstance(gas_prop.fs.unit.dens_mol_comp, Var)
    assert isinstance(gas_prop.fs.unit.dens_mass, Var)
    assert isinstance(gas_prop.fs.unit.cp_mol_comp, Var)
    assert isinstance(gas_prop.fs.unit.cp_mol, Var)
    assert isinstance(gas_prop.fs.unit.cp_mass, Var)
    assert isinstance(gas_prop.fs.unit.visc_d, Var)


@pytest.mark.unit
def test_setInputs_state_block(gas_prop):
    assert degrees_of_freedom(gas_prop.fs.unit) == 0


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(gas_prop):
    initialization_tester(
            gas_prop)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve(gas_prop):

    assert hasattr(gas_prop.fs.unit, "mw")
    assert hasattr(gas_prop.fs.unit, "cp_mol")
    assert hasattr(gas_prop.fs.unit, "enth_mol")

    results = solver.solve(gas_prop)

    # Check for optimal solution
    assert results.solver.termination_condition == \
        TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solution(gas_prop):
    assert (pytest.approx(1, abs=1e-2) ==
            gas_prop.fs.unit.mw.value)
    assert (pytest.approx(1, abs=1e-2) ==
            gas_prop.fs.unit.cp_mol.value)
    assert (pytest.approx(1, abs=1e-2) ==
            gas_prop.fs.unit.enth_mol.value)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for idaes.dmf.commands
"""
import logging
import os
import shutil
import sys

#
import pytest

#
from idaes.dmf import dmfbase, commands, errors, workspace, util
from idaes.util.system import mkdtemp
from .util import init_logging

__author__ = "Dan Gunter"

init_logging()
_log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def wspath():
    dirname = mkdtemp()
    yield dirname
    # teardown
    shutil.rmtree(dirname)


@pytest.mark.unit
def test_workspace_init(wspath):
    commands.workspace_init(wspath, {"some": "metadata"})
    # try again. Should work, since it's OK to init twice
    commands.workspace_init(wspath, {"some": "metadata"})


@pytest.mark.unit
def test_workspace_info(wspath):
    commands.workspace_init(wspath, {"some": "metadata"})
    commands.workspace_info(wspath)

    #    subdir = os.path.join(wspath, 'stuff')
    #    os.mkdir(subdir)
    #    commands.workspace_info(subdir)

    notasubdir = os.path.join(wspath, "nope")
    try:
        commands.workspace_info(notasubdir)
        assert False, "Nonexistent subdir workspace info success"
    except errors.CommandError:
        pass


@pytest.mark.unit
def test_find_html_docs(wspath):
    filedir = os.path.dirname(__file__)
    docpath = os.path.join(filedir, "..", "docs", "build", "html")
    if os.path.exists(docpath):
        commands.workspace_init(wspath, {}, html_paths=[docpath])
        dmfobj = dmfbase.DMF(wspath)
        filenames = commands.find_html_docs(dmfobj, dmfobj)
        assert len(filenames) > 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for DMF CLI
"""
# stdlib
import json
import logging
import os
from pathlib import Path
import sys
import time
from typing import Union

# third-party
from click.testing import CliRunner
import pytest

# package
from idaes.dmf.cli import init, register, info, ls, related, rm, status
from idaes.dmf.dmfbase import DMFConfig, DMF
from idaes.dmf.workspace import Workspace
from idaes.dmf import resource

from . import create_module_scratch, rmtree_scratch

__author__ = "Dan Gunter"

_log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


scratch_path: Union[Path, None] = None
dmf_context_num = 1
# on_windows = sys.platform == "win32"  -- maybe needed later


def setup_module(module):
    global scratch_path
    scratch_path = create_module_scratch(module.__name__)


def teardown_module(module):
    rmtree_scratch(scratch_path)


DATAFILE = "foo.txt"


@pytest.fixture()
def dmf_context():
    """Switch DMF context to a random subdir, then switch back when done.
    """
    global dmf_context_num
    os.chdir(os.path.expanduser("~"))  # make sure we start in HOME
    path = scratch_path / str(dmf_context_num)
    try:
        path.mkdir()
    except:
        pass
    dmf_path = (path / ".dmf").absolute()
    DMFConfig._filename = str(dmf_path)
    origdir = os.getcwd()
    os.chdir(str(path))
    with open(DATAFILE, "w") as fp:
        fp.write("This is some sample data")
    fp.close()
    yield path
    os.unlink(DATAFILE)
    DMFConfig._filename = str(Path("~/.dmf").expanduser())
    os.chdir(origdir)
    dmf_context_num += 1


def create_foo_workspace(runner):
    if (Path("ws") / Workspace.WORKSPACE_CONFIG).exists():
        create_flag = ""
    else:
        create_flag = "--create"
    result = runner.invoke(
        init, ["ws", create_flag, "--name", "foo", "--desc", "foo workspace description"]
    )
    assert result.exit_code == 0
    return result


@pytest.mark.unit()
def test_dmf_find(dmf_context, runner):
    filename = DATAFILE
    create_foo_workspace(runner)
    # register some objects in a workspace
    result = runner.invoke(register, [filename])
    id_all = result.output.strip()
    id_4 = id_all[:4]
    #
    result = runner.invoke(info, ["--no-color", id_4], catch_exceptions=False)
    assert result.exit_code == 0
    assert filename in result.output
    #
    result = runner.invoke(
        info, ["--no-color", "--format", "json", id_4], catch_exceptions=False
    )
    assert result.exit_code == 0
    assert filename in result.output
    out = result.output.strip()
    assert out.startswith("{") and out.endswith("}")
    assert '"relations"' in out
    #
    result = runner.invoke(
        info, ["--no-color", "--format", "jsonc", id_4], catch_exceptions=False
    )
    assert result.exit_code == 0
    assert filename in result.output
    out = result.output.strip()
    j = json.loads(out)
    assert len(j["datafiles"]) == 1


@pytest.mark.unit()
def test_dmf_init(dmf_context, runner):
    create_foo_workspace(runner)
    assert (Path("ws") / Workspace.WORKSPACE_CONFIG).exists()
    #
    result = runner.invoke(init, ["ws2", "--create"], input="foo\nfoo desc\n")
    assert result.exit_code == 0
    assert (Path("ws2") / Workspace.WORKSPACE_CONFIG).exists()
    #
    result = runner.invoke(init, ["doesnotexist"])
    assert result.exit_code != 0
    assert "not found" in result.output
    #
    os.mkdir("some_random_directory")
    result = runner.invoke(init, ["some_random_directory"])
    assert result.exit_code != 0
    #
    result = runner.invoke(init, ["ws", "--create"])
    assert result.exit_code != 0
    assert "exists" in result.output


@pytest.mark.unit()
def test_dmf_ls(dmf_context, runner):
    create_foo_workspace(runner)
    # add some files
    files = [f"foo1{n}" for n in range(5)]
    files.append("bar1")
    for f in files:
        with open(f, "w") as fp:
            fp.write("This is some sample data")
        runner.invoke(register, [f])  # add file to DMF
    # regular
    result = runner.invoke(ls, ["--no-color"])
    assert result.exit_code == 0
    output1 = result.output
    result = runner.invoke(ls, ["--no-color"])
    assert result.output == output1
    # sort
    result = runner.invoke(ls, ["--no-color", "-S", "modified"])
    assert result.exit_code == 0
    output1 = result.output
    result = runner.invoke(ls, ["--no-color", "--sort", "modified"])
    assert result.output == output1


# Due to issue with tempfiles
@pytest.mark.component()
def test_dmf_register(dmf_context, runner):
    create_foo_workspace(runner)
    # create CSV file
    filename = "file.csv"
    with open(filename, "w") as fp:
        fp.write("index,time,value\n1,0.1,1.0\n2,0.2,1.3\n")
    # csv
    result = runner.invoke(register, ["file.csv", "--info"], catch_exceptions=False)
    assert result.exit_code == 0
    assert filename in result.output
    assert "version" in result.output
    # csv again / no-unique
    result = runner.invoke(register, ["file.csv",], catch_exceptions=False)
    assert result.exit_code != 0
    result = runner.invoke(
        register, ["file.csv", "--no-unique"], catch_exceptions=False
    )
    assert result.exit_code == 0
    # json
    not_json = "notreally.json"
    with open(not_json, "w") as fp:
        fp.write("totally bogus\n")
    result = runner.invoke(register, [not_json], catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(
        register, [not_json, "--strict", "--no-unique"], catch_exceptions=False
    )
    assert result.exit_code != 0
    # notebook
    not_nb = "my.ipynb"
    with open(not_nb, "w") as fp:
        fp.write("foo\n")
    result = runner.invoke(register, [not_nb, "-t", "notebook"])
    assert result.exit_code != 0  # error: not JSON
    result = runner.invoke(register, [not_nb, "-t", "data"])
    assert result.exit_code == 0
    # relations
    for text_file in "shoebox", "shoes", "closet":
        open(f"{text_file}.txt", "w")
    result = runner.invoke(register, ["shoebox.txt"], catch_exceptions=False)
    assert result.exit_code == 0
    shoebox_id = result.output.strip()
    result = runner.invoke(
        register, ["shoes.txt", "--contained", shoebox_id], catch_exceptions=False
    )
    assert result.exit_code == 0
    shoe_id = result.output.strip()
    result = runner.invoke(info, [shoe_id, "--format", "jsonc"])
    assert result.exit_code == 0
    # relations-2
    for text_file in "shoebox", "shoes", "closet":
        open(f"{text_file}.txt", "w")
    result = runner.invoke(
        register,
        ["closet.txt", "--is-subject", "--contained", shoebox_id],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    closet_id = result.output.strip()
    result = runner.invoke(info, [shoebox_id, "--format", "jsonc"])
    assert result.exit_code == 0
    data = json.loads(result.output)
    assert len(data["relations"]) == 2
    for rel in data["relations"]:
        if rel["role"] == "subject":
            assert rel["identifier"] == shoe_id
        else:
            assert rel["identifier"] == closet_id


@pytest.mark.unit()
def test_dmf_related(dmf_context, runner):
    create_foo_workspace(runner)
    # add the fully-connected 4 resources
    dmf = DMF()
    rlist = [
        resource.Resource(value={"desc": ltr, "aliases": [ltr], "tags": ["graph"]})
        for ltr in "ABCD"
    ]
    A_id = rlist[0].id  # root resource id, used in testcode
    relation = resource.Predicates.uses
    for r in rlist:
        for r2 in rlist:
            if r is r2:
                continue
            resource.create_relation(r, relation, r2)
    for r in rlist:
        dmf.add(r)
    #
    result = runner.invoke(
        related, [A_id, "--no-unicode", "--no-color"], catch_exceptions=False
    )
    assert result.exit_code == 0
    rlines = result.output.split("\n")
    nrelations = sum(1 for _ in filter(lambda s: resource.Predicates.uses in s, rlines))
    assert nrelations == 12  # 3 blocks of (1 + 3)


@pytest.mark.unit()
def test_dmf_rm(dmf_context, runner):
    create_foo_workspace(runner)
    # add some files named `file{1-5}`
    for i in range(1, 6):
        filename = f"file{i}.txt"
        with open(filename, "w") as fp:
            fp.write(f"file #{i}")
        runner.invoke(register, [filename])
    #
    result = runner.invoke(ls, ["--no-color", "--no-prefix"])
    assert result.exit_code == 0
    output1 = result.output
    output1_lines = output1.split("\n")
    rsrc_id = output1_lines[1].split()[0]  # first token
    result = runner.invoke(rm, [rsrc_id, "--yes", "--no-list"], catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(ls, ["--no-color", "--no-prefix"])
    assert result.exit_code == 0
    output2 = result.output
    output2_lines = output2.split("\n")
    assert len(output2_lines) == len(output1_lines) - 1
    #
    result = runner.invoke(ls, ["--no-color"])
    assert result.exit_code == 0
    output1 = result.output
    output1_lines = output1.split("\n")
    rsrc_id = output1_lines[1].split()[0]  # first token
    result = runner.invoke(rm, [rsrc_id, "--yes", "--no-list"], catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(ls, ["--no-color", "--no-prefix"])
    assert result.exit_code == 0
    output2 = result.output
    output2_lines = output2.split("\n")
    assert len(output2_lines) == len(output1_lines) - 1


@pytest.mark.unit()
def test_dmf_status(dmf_context, runner):
    create_foo_workspace(runner)
    #
    result = runner.invoke(status, ['--no-color'])
    assert result.exit_code == 0
    assert "settings" in result.output
    assert "name: foo" in result.output
    #
    result = runner.invoke(status, ['--no-color', '--show', 'files'])
    assert result.exit_code == 0
    assert "settings" in result.output
    assert "name: foo" in result.output
    assert "files:" in result.output
    #
    result = runner.invoke(status, ['--no-color', '--show', 'files',
        '--show', 'htmldocs'])
    assert result.exit_code == 0
    assert "settings" in result.output
    assert "name: foo" in result.output
    assert "html" in result.output
    #
    result = runner.invoke(status, ['--no-color', '-a'])
    assert result.exit_code == 0
    assert "settings" in result.output
    assert "name: foo" in result.output
    assert "html" in result.output
    assert "logging:" in result.output



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for common methods used by generic framework

Author: A Lee
"""
import pytest

from types import MethodType

from idaes.generic_models.properties.core.generic.utility import (
    GenericPropertyPackageError, get_method, get_phase_method,
    get_component_object, get_bounds_from_config)

from pyomo.environ import Block, units as pyunits
from pyomo.common.config import ConfigBlock, ConfigValue
from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError


@pytest.fixture
def frame():
    m = Block(concrete=True)
    m.params = Block()
    m.params.config = ConfigBlock()
    m.params.config.declare("test_arg", ConfigValue())
    m.params.config.declare("state_bounds", ConfigValue())

    def get_component(self, comp):
        return getattr(self, comp)
    m.params.get_component = MethodType(get_component, m.params)
    m.params.get_phase = MethodType(get_component, m.params)

    m.params.comp = Block()
    m.params.comp.config = ConfigBlock()
    m.params.comp.config.declare("test_arg_2", ConfigValue())

    return m


@pytest.mark.unit
def test_generic_property_package_error():
    with pytest.raises(
            PropertyPackageError,
            match="Generic Property Package instance block called for "
            "prop, but was not provided with a method "
            "for this property. Please add a method for this property "
            "in the property parameter configuration."):
        raise GenericPropertyPackageError("block", "prop")


@pytest.mark.unit
def test_get_component_object(frame):
    assert get_component_object(frame, "comp") is frame.params.comp


class TestGetMethod():
    @pytest.mark.unit
    def test_get_method_invalid_name(self, frame):
        with pytest.raises(
                AttributeError,
                match="ScalarBlock Generic Property Package called for "
                "invalid configuration option foo. Please contact the "
                "developer of the property package."):
            get_method(frame, "foo")

    @pytest.mark.unit
    def test_get_method_none(self, frame):
        with pytest.raises(
                GenericPropertyPackageError,
                match="Generic Property Package instance ScalarBlock "
                "called for test_arg, but was not provided with a "
                "method for this property. Please add a method for "
                "this property in the property parameter "
                "configuration."):
            get_method(frame, "test_arg")

    @pytest.mark.unit
    def test_get_method_not_callable(self, frame):
        frame.params.config.test_arg = "foo"
        with pytest.raises(
                ConfigurationError,
                match="ScalarBlock Generic Property Package received "
                "invalid value for argument test_arg. Value must be a "
                "method, a class with a method named expression or a "
                "module containing one of the previous."):
            get_method(frame, "test_arg")

    @pytest.mark.unit
    def test_get_method_simple(self, frame):
        def test_arg():
            return "bar"

        frame.params.config.test_arg = test_arg

        mthd = get_method(frame, "test_arg")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_class_w_method(self, frame):
        class TestClass():
            def test_arg():
                return "bar"

        frame.params.config.test_arg = TestClass

        mthd = get_method(frame, "test_arg")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_class_w_return_expression(self, frame):
        class TestClass():
            def return_expression(*args, **kwargs):
                return "bar"

        frame.params.config.test_arg = TestClass

        mthd = get_method(frame, "test_arg")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_phase(self, frame):
        def test_arg():
            return "bar"

        frame.params.config.test_arg = {"test_phase": test_arg}

        mthd = get_method(frame, "test_arg", phase="test_phase")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_comp(self, frame):
        def test_arg():
            return "bar"

        frame.params.comp.config.test_arg_2 = test_arg

        mthd = get_method(frame, "test_arg_2", comp="comp")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_phase_comp(self, frame):
        def test_arg():
            return "bar"

        frame.params.comp.config.test_arg_2 = {"test_phase": test_arg}

        mthd = get_method(frame, "test_arg_2", comp="comp", phase="test_phase")
        assert mthd() == "bar"


class TestGetPhaseMethod():
    @pytest.mark.unit
    def test_get_phase_method_invalid_name(self, frame):
        with pytest.raises(
                AttributeError,
                match="ScalarBlock Generic Property Package called for "
                "invalid configuration option foo. Please contact the "
                "developer of the property package."):
            get_phase_method(frame, "foo", "comp")

    @pytest.mark.unit
    def test_get_phase_method_none(self, frame):
        with pytest.raises(
                GenericPropertyPackageError,
                match="Generic Property Package instance ScalarBlock "
                "called for test_arg_2, but was not provided with a "
                "method for this property. Please add a method for "
                "this property in the property parameter "
                "configuration."):
            get_phase_method(frame, "test_arg_2", "comp")

    @pytest.mark.unit
    def test_get_phase_method_not_callable(self, frame):
        frame.params.comp.config.test_arg_2 = "foo"
        with pytest.raises(
                ConfigurationError,
                match="ScalarBlock Generic Property Package received "
                "invalid value for argument test_arg_2. Value must be a "
                "method, a class with a method named expression or a "
                "module containing one of the previous."):
            get_phase_method(frame, "test_arg_2", "comp")

    @pytest.mark.unit
    def test_get_phase_method_simple(self, frame):
        def test_arg():
            return "bar"

        frame.params.comp.config.test_arg_2 = test_arg

        mthd = get_phase_method(frame, "test_arg_2", "comp")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_phase_method_class_w_method(self, frame):
        class TestClass():
            def test_arg_2():
                return "bar"

        frame.params.comp.config.test_arg_2 = TestClass

        mthd = get_phase_method(frame, "test_arg_2", "comp")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_phase_method_class_w_return_expression(self, frame):
        class TestClass():
            def return_expression():
                return "bar"

        frame.params.comp.config.test_arg_2 = TestClass

        mthd = get_phase_method(frame, "test_arg_2", "comp")
        assert mthd() == "bar"


class TestGetBoundsFromConfig():
    @pytest.mark.unit
    def test_no_state_bounds(self, frame):
        bounds, value = get_bounds_from_config(frame, "foo", "bar")

        assert bounds == (None, None)
        assert value is None

    @pytest.mark.unit
    def test_no_state_bounds_3(self, frame):
        frame.params.config.state_bounds = {
            "test_state": (1, 2, 3)}
        bounds, value = get_bounds_from_config(frame, "test_state", "bar")

        assert bounds == (1, 3)
        assert value == 2

    @pytest.mark.unit
    def test_no_state_bounds_4(self, frame):
        frame.params.config.state_bounds = {
            "test_state": (1, 2, 3, pyunits.km)}
        bounds, value = get_bounds_from_config(frame, "test_state", pyunits.m)

        assert bounds == (1000, 3000)
        assert value == 2000
